from vector import Vector
import math
import random
from pygame.draw import lines as drawlines
import pygame.sprite
import pygame.surface

LINK_DISTANCE = 5
SPRING_K = 0.00001
GRAVITY_MAGNITUDE = 0.0002
LINK_GRAVITY = Vector(0.0, GRAVITY_MAGNITUDE)
HANG_GRAVITY_MAGNITUDE = 0.02
HANG_GRAVITY = Vector(0.0, HANG_GRAVITY_MAGNITUDE)
VINE_COLOR = [0, 60, 0]

class Vine(pygame.sprite.DirtySprite):
    def __init__(self, attachpoint, length):
        pygame.sprite.DirtySprite.__init__(self)
        self.links = []
        self.attachpoint = VineLink(attachpoint, 1.0)
        self.attachpoint_modifier = 3
        self.last_attachpoint_modify = 0
        self.attachpoint_modify_threshold = random.randint(500, 1500)
        self.length = length
        
        self.hang_link = None
        self.hang_links_free = None
        self.hang_link_index = None
        
        link_amount = length // LINK_DISTANCE
        
        for i in range(link_amount):
            newlink = VineLink(Vector(attachpoint.x, attachpoint.y + (i+1)*LINK_DISTANCE), 5.0-4.0*i/link_amount)
            self.links.append(newlink)
        
        self.free = True
        
        self.image = pygame.Surface((length*2, length))
        self.image.convert()
        self.image.set_colorkey([0,0,0])
        self.rect = self.image.get_rect()
        self.rect.topleft = (self.attachpoint.position.x-length, self.attachpoint.position.y)
        self.dirty = 2
    
    def update(self, dt, world):
        if self.free:
            self.last_attachpoint_modify += dt
            if self.last_attachpoint_modify > self.attachpoint_modify_threshold:
                self.attachpoint.position.x += self.attachpoint_modifier
                self.attachpoint_modifier = -self.attachpoint_modifier
                self.last_attachpoint_modify = 0
                
            for i in range(len(self.links)):
                if i == 0:
                    prevlink = self.attachpoint
                else:
                    prevlink = self.links[i-1]
                if i == len(self.links)-1:
                    nextlink = None
                else:
                    nextlink = self.links[i+1]
                                
                self.links[i].update(dt, prevlink, nextlink)
        else:
            self.hang_link.update(dt)
            for i in range(len(self.hang_links_free)):
                if i == 0:
                    prevlink = self.hang_link
                else:
                    prevlink = self.hang_links_free[i-1]
                if i == len(self.hang_links_free)-1:
                    nextlink = None
                else:
                    nextlink = self.hang_links_free[i+1]
                
                self.hang_links_free[i].update(dt, prevlink, nextlink)
        self.draw(self.image)
        return
    
    def draw(self, dest):
        dest.fill([0,0,0])   
        pointlist = []
        pointlist.append(self.get_relative_coords(self.attachpoint.position))
        if self.free:
            for i in range(len(self.links)):
                pointlist.append(self.get_relative_coords(self.links[i].position))
            
        else:
            pointlist.append(self.get_relative_coords(self.hang_link.position))
            for i in range(len(self.hang_links_free)):
                pointlist.append(self.get_relative_coords(self.hang_links_free[i].position))
        drawlines(dest, VINE_COLOR, False, pointlist, 5)
        return
    
    def get_relative_coords(self, coords):
        x = coords.x - self.rect.left
        y = coords.y - self.rect.top
        return [x, y]
    
    def catch(self, link, player):
        player.states["swinging_left"].attached_vine = self
        player.states["swinging_right"].attached_vine = self
        self.free = False
        self.hang_link_index = self.links.index(link)
        self.hang_link = HangLink(link.position, self.attachpoint, self.hang_link_index * LINK_DISTANCE)
        self.hang_link.catch(player)
        self.hang_links_free = []
        for i in range(self.links.index(link)+1, len(self.links)):
            self.hang_links_free.append(self.links[i])
        
        return self.hang_link
    
    def let_go(self, player):
        self.free = True
        for i in range(0, self.hang_link_index):
            self.links[i].position = self.attachpoint.position + self.hang_link.diff * LINK_DISTANCE
        for i in range(self.hang_link_index, len(self.links)):
            self.links[i].position = self.hang_link.position + Vector(0, (i-self.hang_link_index) * LINK_DISTANCE)
        self.hang_link = None
        self.hang_links_free = None
    
class VineLink:
    def __init__(self, position, mass = 1.0, max_distance = LINK_DISTANCE):
        self.position = position
        self.mass = mass
        self.velocity = Vector(0.0, 0.0)
        self.max_distance = max_distance          
    
    def update(self, dt, prevlink, nextlink):
        acceleration = (self.get_force_between_links(prevlink) + self.get_force_between_links(nextlink) + LINK_GRAVITY) / self.mass
        self.velocity += acceleration * dt
        self.velocity = self.velocity.trunc(5)
        self.position += self.velocity * dt
        diff = prevlink.position - self.position
        if diff.magnitude_sq() > ((self.max_distance) ** 2):
            self.position = prevlink.position - diff.trunc(self.max_distance)
        
    def get_force_between_links(self, otherlink):
        if otherlink == None:
            return Vector(0.0, 0.0)
        
        difference = otherlink.position - self.position
        
        x = difference.magnitude() - self.max_distance
        direction = difference.normalize()
        return SPRING_K * x * direction
        
class HangLink:
    def __init__(self, position, attachpoint, r):
        self.attachpoint = attachpoint
        self.position = position
        self.velocity = Vector(0.0,0.0)
        self.r = r
        self.diff = (self.attachpoint.position - self.position).normalize()
        self.perp = self.diff.perp()
        self.mass = 20
    
    def update(self, dt):
        self.diff = (self.attachpoint.position - self.position).normalize()
        self.perp = self.diff.perp()
        
        if self.attachpoint.position.x <= self.position.x:
            self.perp = -self.perp
        
        #get the normal force
        cos_diff_grav = self.diff.dot(-HANG_GRAVITY)/HANG_GRAVITY_MAGNITUDE 
        acceleration = (HANG_GRAVITY + self.diff * cos_diff_grav * HANG_GRAVITY_MAGNITUDE) / self.mass
        self.velocity += acceleration * dt
        #align velocity with the tangent
        vel_magn = self.velocity.magnitude()
        cos_perp_vel = self.perp.dot(self.velocity)/(1*vel_magn)
        self.velocity = cos_perp_vel * self.perp * vel_magn
        
        self.position += self.velocity * dt
    
    def catch(self, player):
        player.states["swinging_left"].attached_vinelink = self
        player.states["swinging_right"].attached_vinelink = self
        vel_magn = player.velocity.magnitude()
        cos_perp_velocity = self.perp.dot(player.velocity)/(self.perp.magnitude()*vel_magn)
        self.velocity = self.perp * cos_perp_velocity * vel_magn